/*
   MacComm.cpp

   Interface til JVC's Mac motorer.


   Setup brug MacTalk:

   Tilslut een motor af gangen

   I Main:
      St Startup mode: Position
      St Communication: 2 for X, 3 for y, 4 for Z etc (akse handle)
      St In position window: 10

   I Filter Setup
      St Current Filter: Soft(Low noise, max 3000 RPM)
      St Position filter: Speed Responce=Slow, Hardness=Hard
      St Follow error compensation: Type=Dynamic, Degree=0%

   Tryk p "Save in Flash"


   SCAN technology ApS
   Claus Dybdal Nielsen

*/
//---------------------------------------------------------------------------

/***** proto  - ikke i brug
#define MAX_AXIS 10
typedef struct
{
   int ComPort;                     // RS232 port nummer 1,2,3...
   int nAxis;                       // antal akser (Axis=0,1,2,3...)
   double Pitch[MAX_AXIS];          // flyttet lngde per omgang af motor
   double HomeSpeed[MAX_AXIS];      // Speed under homing og stroke med fortegn
   double HomeTorquePct[MAX_AXIS];  // Moment ved sgning af Home og Stroke
   double HomeDist[MAX_AXIS];       // afstand fra home/stroke til start/slut
}
   MAC_INTERFACE;
**************************/

#pragma hdrstop

#include <Types.h>
#include "MACCOMMLib_OCX.h"
#include "MacComm.h"


//--- Mac Defines --------------------------------------------------------------
#define ALL             255
#define MODE_REG        2
#define HOME_TYPE1      13  /* kun en vej */
#define HOME_TYPE2      14  /* frem/tilbage */
#define HOME1_MODE      12  /* Torque mode */
#define ERR_STAT        35
#define IN_POSITION     0x10
#define V_HOME          40
#define T_HOME          41
#define T_SOLL          7
#define P_HOME          38
#define MODE_POS        2
#define STARTMODE       37
#define MODE_PASSIVE    0
#define INPOSWIN        33
#define MYADDR          156   /* bruges ved prog af akse addr */
#define P_IST           10    /* aktuel position */
#define P_SOLL          3     /* hvor motor skal kre til */
#define V_IST           12    /* ls speed (m ikke overskrives) */
#define V_SOLL          5     /* st max speed */
//--- end Mac Defines ----------------------------------------------------------

#define MAX_AXIS 10        /* max akser tilsluttet (+2 pga addr starter ved 2*/

PRIVATE bool TorqueHome=true; // ellers switch
// 100% torque = 1023
#define TORQUE_PCT(pct)   pct*10.23




// omregn m/min til p/sec
#define mMin2PulsSec(Axis,v)  16*4096*1000*v/Pitch[Axis]/520.8/60


PRIVATE int nAxis;           // antal akser tilsluttet
EXPORT const int Xaxis =0;  // axis handle
EXPORT const int Yaxis =1;

PRIVATE bool IsInit=false;
PRIVATE const double nPulses=4096; // MacMotor encoder pulses/omg
PRIVATE double Pitch[MAX_AXIS];// [mm/omg]
PRIVATE double Gear[MAX_AXIS];
PRIVATE double HomeSpeed[MAX_AXIS]; // m/min
PRIVATE double HomePos[MAX_AXIS]; // [mm]
PRIVATE double Max[MAX_AXIS]; // [mm]


//---------------------------------------------------------------------------

#pragma package(smart_init)
PRIVATE TMacComm *Mac;

PRIVATE bool Mac_Read( const short int Axis, const short int ParameterNo, int& ReadValue )
{
   // iflge Lars/JVL tager en Read ca 15 mS
   // og en Write ca 10 mS (baud 19.2K)

   if(!IsInit) return true;

   long NormValue;

   // Addr= Axis+2 pnr ved Axis=255 => ALL axis
   int Addr= (Axis<255)?Axis+2:255;
   if(!Mac->ReadParameter(Addr, ParameterNo, &NormValue))
      {
      String Err=Mac->GetLastErrorStr(Mac->GetLastError());
      Application->MessageBox(Err.c_str(),"MacErr",0);
      return true;
      }

   switch(Mac->GetParameterType(ParameterNo))
      {
      case 0:
      case 1:
      case 2:
      default: ReadValue = NormValue;      break;  // as is
      case 3:  ReadValue = NormValue/4096; break;  // Fixed 4
      case 4:  ReadValue = NormValue/256;  break;  // Fixed 8
      }

   return false;
}

PRIVATE bool Mac_Write( const short int Axis, const short int ParameterNo, const long SendValue )
{
   if(!IsInit) return true;
   long LocalValue;

   switch(Mac->GetParameterType(ParameterNo))
      {
      case 0:
      case 1:
      case 2:
      default:LocalValue=SendValue;      break;
      case 3: LocalValue=SendValue*4096; break;
      case 4: LocalValue=SendValue*256; break;
      }

   // Addr= Axis+2 pnr ved Axis=255 => ALL axis
   int Addr= (Axis<255)?Axis+2:255;
  if(!Mac->WriteParameter(Addr,ParameterNo,LocalValue))
      {
      String Err=Mac->GetLastErrorStr(Mac->GetLastError());
      Application->MessageBox(Err.c_str(),"MacErr",0);
      return true;
      }
  return false;
}


//--- Init/Term ----------------------------------------------------------------
EXPORT bool Mac_Init(TMacComm *MacObj,
                     const int ComPort,   // COM1,COM2...
                     const double xPitch,  // mm/omgang
                     const double yPitch,
                     const double xHomeSpeed, // m/min med fortegn
                     const double yHomeSpeed)
{
   nAxis=2;
   Mac=MacObj;

   Mac->ComPort=ComPort;

   if(!Mac->OpenPort())
      {
      Application->MessageBox("Can't open communication to servo!","OCX error",0);
      return true;
      }

   long Mode; // Xaddr=2, Yaddr=3...

   for(int Addr=2;Addr<nAxis+2;Addr++)
      {
      if(!Mac->ReadParameter(Addr,MODE_REG,&Mode))
         {
         char Axis='X'+Addr-2;
         String Text="No reply from "+String(Axis)+"-axis\n\nPlease check cabling and power";
         Application->MessageBox(Text.c_str(),"Startup error",0);
         return true;
         }
      // Hvis akse ikke er i Positions mode:
      if(Mode!=MODE_POS) Mac->WriteParameter(Addr,MODE_REG,MODE_POS);
      }

   IsInit=true;

   // br gres med loop under forudstning af X=0, Y=1...N=n

   //--- X:
   Pitch[Xaxis]=xPitch; // fx 10 mm spindelstigning
   Gear[Xaxis]= nPulses/Pitch[Xaxis];  // pulser/mm fra encoder
   HomeSpeed[Xaxis]=mMin2PulsSec(Xaxis,xHomeSpeed);
   HomePos[Xaxis]= ROUND(-3*Gear[Xaxis]);// samme fortegn som vhome  [pulser]!!!
   Max[Xaxis]=200; // DB: loades fra fil senere


   //--- Y:
   Pitch[Yaxis]=yPitch;
   Gear[Yaxis]= nPulses/Pitch[Yaxis];
   HomeSpeed[Yaxis]=mMin2PulsSec(Yaxis,yHomeSpeed);
   HomePos[Yaxis]= ROUND(-3*Gear[Yaxis]);// samme fortegn som vhome
   Max[Yaxis]=200; // DB: loades fra file senere

   return false;
}

EXPORT void Mac_Term(void)
{
   if(!IsInit) return;

   // St alle akse hastigheder til 0, s momentet bevares (servo on):
   // Hvis man stter til "Passive" mode, kan der komme en positioneringsfejl p op til 30 pulser.
   for(int Axis=0;Axis<nAxis;Axis++)
      Mac_Write( Axis, V_SOLL,0 );
   Mac->ClosePort();
   IsInit=false;
}
//------------------------------------------------------------------------------


PRIVATE bool Mac_WriteFlash(void)
{
   if(!IsInit) return true;

   // St P_SOLL til 0 for s der ikke kres til en position
   // nr motor tndes igen (!)

   Mac_Write(ALL,P_SOLL,0);

   if(!Mac->WriteToFlash(ALL))
      {
      String Err=Mac->GetLastErrorStr(Mac->GetLastError());
      Application->MessageBox(Err.c_str(),"MacErr",0);
      return true;
      }
  Mac->ResetWait(ALL);
  return false;
}


//--- Set ----------------------------------------------------------------------

EXPORT void Mac_SetMaxRPM(const int Axis, const double rpm)
{
   if(!IsInit) return;

   // rpm= 10..4000
   double RPM=rpm;
   RPM=RPM>4000?4000:RPM;
   RPM=RPM<0?10:RPM;

   int v_soll=16*4096*RPM/520.8/60;
   Mac_Write(Axis,V_SOLL,v_soll);
}

EXPORT void Mac_SetMaxSpeed(const int Axis, const double Speed) // [m/min]
{
   if(!IsInit) return;
   Mac_SetMaxRPM(Axis,1000*Speed/Pitch[Axis]); // omregn til RPM
}

//--- Get ----------------------------------------------------------------------

EXPORT double Mac_Get(const int Axis)
{
   if(!IsInit) return 0;
   int mPos;
   Mac_Read(Axis,P_IST,mPos);
   return mPos/Gear[Axis]; // [mm]
}

EXPORT double Mac_GetRPM(const int Axis)
{
   int V;
   Mac_Read(Axis,V_IST,V);
   double rpm=(520.8*V)*60/16/4096;
   rpm*=16; // faktor 16 forskel mellem V_IST og V_SOLL (?)
   return floor(rpm*10)/10; // 1 ciffer
}

EXPORT double Mac_GetSpeed(const int Axis) // [m/min]
{
   if(!IsInit) return 0;
   return Mac_GetRPM(Axis)*Pitch[Axis]*0.001; // omregn fra RPM
}

//--- Flags --------------------------------------------------------------------
EXPORT bool Mac_IsInit(void) {return IsInit;}

EXPORT bool Mac_InMotion(const int Axis)
{
   int V;
   Mac_Read(Axis,V_IST,V);
   return ABS(V)>2;
}

//--- Position -----------------------------------------------------------------
EXPORT void Mac_Goto(const int Axis, const double Pos)
{
   if(!IsInit) return;
   int mPos=ROUND(Pos*Gear[Axis]);
   Mac_Write(Axis,P_SOLL,mPos);
}

EXPORT void Mac_GotoMax(const int Axis)
{
   if(!IsInit) return;
   int mPos=ROUND(Max[Axis]*Gear[Axis]);
   Mac_Write(Axis,P_SOLL,mPos);
}

EXPORT void Mac_StopAll(void)
{
   if(!IsInit) return;

   int Pos[MAX_AXIS];
   int RPM[MAX_AXIS];
   bool InMotion;

   //-- Gem speed for alle akser
   for(int Axis=0;Axis<nAxis;Axis++)
      Mac_Read(Axis,V_SOLL,RPM[Axis]);

   //--- stands alle akser (speed=0)
   for(int Axis=0;Axis<nAxis;Axis++)
      Mac_Write(Axis,V_SOLL,0);

   //Mac_Write(ALL,V_SOLL,0);  duer ikke! skal skrives til hver enkelt akse!

   do{
      InMotion=false;
      for(int Axis=0;Axis<nAxis;Axis++)
         if(Mac_InMotion(Axis))
            InMotion=true;
      }
         while(InMotion);

   //--- Genetabler akser:
   for(int Axis=0;Axis<nAxis;Axis++)
      {
      Mac_Read(Axis,P_IST,Pos[Axis]);    // ls nuvrende P_IST pos
      Mac_Write(Axis,P_SOLL,Pos[Axis]);  // skriv pos i P_SOLL
      Mac_Write(Axis,V_SOLL,RPM[Axis]);  // skriv RPM tilbage i V_SOLL
      }
}


EXPORT void Mac_Stop(const int Axis)
{
   if(!IsInit) return;
   int Pos,Speed;

   //--- Gem Speed og Stop akse:
   Mac_Read(Axis,V_SOLL,Speed);  // ls max speed
   Mac_Write(Axis,V_SOLL,0);     // stands akse (speed=0)

   while(Mac_InMotion(Axis));    // vent til den str stille

   //--- Genetabler:
   Mac_Read(Axis,P_IST,Pos);     // ls nuvrende P_IST pos
   Mac_Write(Axis,P_SOLL,Pos);   // skriv pos i P_SOLL
   Mac_Write(Axis,V_SOLL,Speed);  // skriv speed tilbage i V_SOLL
}

PRIVATE void Mac_HomeTorque(const int Axis)
{
   // Torque home!

   int Val;

   if(Mac_InMotion(Axis))
      Mac_Stop(Axis);

   Mac_Write(Axis,V_HOME, HomeSpeed[Axis]);
   Mac_Write(Axis,T_HOME, TORQUE_PCT(10));
   Mac_Write(Axis,P_HOME, HomePos[Axis]); // afstand m vhome

   Mac_Write(Axis,P_SOLL,0);
   Mac_Write(Axis,MODE_REG,HOME1_MODE);

   do {Mac_Read(Axis,ERR_STAT,Val);}
      while (!(Val & IN_POSITION));

   Application->MessageBox("Home done","Message",0);
}

EXPORT void Mac_HomeTorqueXY(void)
{
   // Torque home!

   //--- Stop XY
   if(Mac_InMotion(Xaxis))Mac_Stop(Xaxis);
   if(Mac_InMotion(Yaxis))Mac_Stop(Yaxis);

   //--- Setup X
   Mac_Write(Xaxis,V_HOME, HomeSpeed[Xaxis]);
   Mac_Write(Xaxis,T_HOME, TORQUE_PCT(10));
   Mac_Write(Xaxis,P_HOME, HomePos[Xaxis]); // afstand m vhome

   //--- Setup Y
   Mac_Write(Yaxis,V_HOME, HomeSpeed[Yaxis]);
   Mac_Write(Yaxis,T_HOME, TORQUE_PCT(10));
   Mac_Write(Yaxis,P_HOME, HomePos[Yaxis]); // afstand m vhome

   //--- Start homeing
   Mac_Write(Xaxis,P_SOLL,0);
   Mac_Write(Xaxis,MODE_REG,HOME1_MODE);

   Mac_Write(Yaxis,P_SOLL,0);
   Mac_Write(Yaxis,MODE_REG,HOME1_MODE);

   int X,Y;
   do {
      Mac_Read(Xaxis,ERR_STAT,X);
      Mac_Read(Yaxis,ERR_STAT,Y);
      }
      while (!((X&IN_POSITION) && (Y&IN_POSITION)));

   Application->MessageBox("Home done XY","Message",0);
}

PRIVATE void Mac_HomeSwitch(const int Axis)
{
   int Val;

   if(Mac_InMotion(Axis))
      Mac_Stop(Axis);

   Mac_Write(Axis,V_HOME, HomeSpeed[Axis]);
   Mac_Write(Axis,P_SOLL,0);  // er ndvendigt for at undg movement efter home!

   Mac_Write(Axis,MODE_REG,HOME_TYPE2);


   // INPOSWIN skal >5 for at virke. Sat til 10
   do {Mac_Read(Axis,ERR_STAT,Val);}
      while (!(Val & IN_POSITION));

   Application->MessageBox("Home done","Message",0);
}

EXPORT void Mac_Home(const int Axis)
{
   if(TorqueHome)
      Mac_HomeTorque(Axis);
   else
      Mac_HomeSwitch(Axis);
}


EXPORT void Mac_FindStroke(const int Axis)
{
   int t_soll,v_soll;

   Mac_Read(Axis,T_SOLL, t_soll);
   Mac_Write(Axis,T_SOLL, TORQUE_PCT(10));
   Mac_Read(Axis,V_SOLL,v_soll);
   Mac_Write(Axis,V_SOLL,HomeSpeed[Axis]);

   Mac_Goto(Axis, 100000);

   while(!Mac_InMotion(Axis)); // vent til den starter
   while(Mac_InMotion(Axis)); // vent til den stopper

   Max[Axis]=Mac_Get(Axis) - 3; // mm
   Mac_Goto(Axis,Max[Axis]);

   Mac_Write(Axis,T_SOLL, t_soll);
   Mac_Write(Axis,V_SOLL,v_soll);
}

EXPORT void Mac_FindStrokeXY(void)
{
   int t_sollX,v_sollX;
   int t_sollY,v_sollY;

   //--- Setup X
   Mac_Read(Xaxis,T_SOLL, t_sollX);
   Mac_Write(Xaxis,T_SOLL, TORQUE_PCT(10));
   Mac_Read(Xaxis,V_SOLL,v_sollX);
   Mac_Write(Xaxis,V_SOLL,HomeSpeed[Xaxis]);

   //--- Setup Y
   Mac_Read(Yaxis,T_SOLL, t_sollY);
   Mac_Write(Yaxis,T_SOLL, TORQUE_PCT(10));
   Mac_Read(Yaxis,V_SOLL,v_sollY);
   Mac_Write(Yaxis,V_SOLL,HomeSpeed[Yaxis]);

   //--- Go XY
   Mac_Goto(Xaxis, 100000);
   Mac_Goto(Yaxis, 100000);

   while(!Mac_InMotion(Xaxis) && !Mac_InMotion(Yaxis)); // vent til de begge er startet

   bool Xdone=false;
   bool Ydone=false;
   do{
      if(!Xdone && !Mac_InMotion(Xaxis))
         {
         Max[Xaxis]=Mac_Get(Xaxis) - 3; // mm
         Mac_Goto(Xaxis,Max[Xaxis]);
         Xdone=true;
         }

      if(!Ydone && !Mac_InMotion(Yaxis))
         {
         Max[Yaxis]=Mac_Get(Yaxis) - 3; // mm
         Mac_Goto(Yaxis,Max[Yaxis]);
         Ydone=true;
         }
      }
   while(!(Xdone && Ydone));

   //--- Genetabler XY
   Mac_Write(Xaxis,T_SOLL, t_sollX);
   Mac_Write(Xaxis,V_SOLL,v_sollX);
   Mac_Write(Yaxis,T_SOLL, t_sollY);
   Mac_Write(Yaxis,V_SOLL,v_sollY);
}


